@@ -2,13 +2,16 @@
 #this repository contains the full copyright notices and license terms.
 from dateutil.relativedelta import relativedelta
 from trytond.model import ModelView, ModelSQL, fields
-from trytond.wizard import Wizard, StateView, StateTransition, Button
+from trytond.wizard import Wizard, StateView, StateTransition, StateAction, \
+    Button
 from trytond.tools import datetime_strftime
-from trytond.pyson import Eval, If
+from trytond.pyson import Eval, If, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
-__all__ = ['FiscalYear', 'CloseFiscalYearStart', 'CloseFiscalYear']
+__all__ = ['FiscalYear',
+    'BalanceNonDeferralStart', 'BalanceNonDeferral',
+    'CloseFiscalYearStart', 'CloseFiscalYear']
 
 STATES = {
     'readonly': Eval('state') == 'close',
@@ -45,32 +48,28 @@ class FiscalYear(ModelSQL, ModelView):
     company = fields.Many2One('company.company', 'Company', required=True,
         domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
+                Eval('context', {}).get('company', -1)),
             ], select=True)
 
     @classmethod
     def __setup__(cls):
         super(FiscalYear, cls).__setup__()
-        cls._constraints += [
-            ('check_dates', 'fiscalyear_overlaps'),
-            ('check_post_move_sequence', 'different_post_move_sequence'),
-        ]
         cls._order.insert(0, ('start_date', 'ASC'))
         cls._error_messages.update({
-            'change_post_move_sequence': 'You can not change ' \
-                    'the post move sequence',
-            'no_fiscalyear_date': 'No fiscal year defined for this date!',
-            'fiscalyear_overlaps':
-                'You can not have 2 fiscal years that overlaps!',
-            'different_post_move_sequence':
-                'You must have different post move sequence per fiscal year!',
-            'account_balance_not_zero':
-                'The balance of the account "%s" must be zero!',
-            'close_error': 'You can not close a fiscal year until ' \
-                    'there is older fiscal year opened!',
-            'reopen_error': 'You can not reopen a fiscal year until ' \
-                    'there is more recent fiscal year closed!',
-            })
+                'change_post_move_sequence': ('You can not change the post '
+                    'move sequence in fiscal year "%s".'),
+                'no_fiscalyear_date': 'No fiscal year defined for "%s".',
+                'fiscalyear_overlaps': ('Fiscal year "%(first)s" and '
+                    '"%(second)s" overlap.'),
+                'different_post_move_sequence': ('Fiscal year "%(first)s" and '
+                    '"%(second)s" have the same post move sequence.'),
+                'account_balance_not_zero': ('The balance of the account "%s" '
+                    'must be zero.'),
+                'close_error': ('You can not close fiscal year "%s" until you '
+                    'close all previous fiscal years.'),
+                'reopen_error': ('You can not reopen fiscal year "%s" until '
+                    'you reopen all later fiscal years.'),
+                })
         cls._buttons.update({
                 'create_period': {
                     'invisible': ((Eval('state') != 'open')
@@ -96,46 +95,63 @@ class FiscalYear(ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    @classmethod
+    def validate(cls, years):
+        super(FiscalYear, cls).validate(years)
+        for year in years:
+            year.check_dates()
+            year.check_post_move_sequence()
+
     def check_dates(self):
         cursor = Transaction().cursor
-        cursor.execute('SELECT id ' \
-                'FROM ' + self._table + ' ' \
-                'WHERE ((start_date <= %s AND end_date >= %s) ' \
-                        'OR (start_date <= %s AND end_date >= %s) ' \
-                        'OR (start_date >= %s AND end_date <= %s)) ' \
-                    'AND company = %s ' \
-                    'AND id != %s',
-                (self.start_date, self.start_date,
-                    self.end_date, self.end_date,
-                    self.start_date, self.end_date,
-                    self.company.id, self.id))
-        if cursor.fetchone():
-            return False
-        return True
+        table = self.__table__()
+        cursor.execute(*table.select(table.id,
+                where=(((table.start_date <= self.start_date)
+                        & (table.end_date >= self.start_date))
+                    | ((table.start_date <= self.end_date)
+                        & (table.end_date >= self.end_date))
+                    | ((table.start_date >= self.start_date)
+                        & (table.end_date <= self.end_date)))
+                & (table.company == self.company.id)
+                & (table.id != self.id)))
+        second_id = cursor.fetchone()
+        if second_id:
+            second = self.__class__(second_id[0])
+            self.raise_user_error('fiscalyear_overlaps', {
+                    'first': self.rec_name,
+                    'second': second.rec_name,
+                    })
 
     def check_post_move_sequence(self):
-        if self.search([
-                    ('post_move_sequence', '=', self.post_move_sequence.id),
-                    ('id', '!=', self.id),
-                    ]):
-            return False
-        return True
+        years = self.search([
+                ('post_move_sequence', '=', self.post_move_sequence.id),
+                ('id', '!=', self.id),
+                ])
+        if years:
+            self.raise_user_error('different_post_move_sequence', {
+                    'first': self.rec_name,
+                    'second': years[0].rec_name,
+                    })
 
     @classmethod
-    def write(cls, fiscalyears, vals):
-        if vals.get('post_move_sequence'):
-            for fiscalyear in fiscalyears:
-                if fiscalyear.post_move_sequence and \
-                        fiscalyear.post_move_sequence.id != \
-                        vals['post_move_sequence']:
-                    cls.raise_user_error('change_post_move_sequence')
-        vals = vals.copy()
-        if 'periods' in vals:
-            operator = ['delete', 'unlink_all', 'unlink', 'create', 'write',
-                    'add', 'set']
-            vals['periods'].sort(
-                lambda x, y: cmp(operator.index(x[0]), operator.index(y[0])))
-        super(FiscalYear, cls).write(fiscalyears, vals)
+    def write(cls, *args):
+        actions = iter(args)
+        args = []
+        for fiscalyears, values in zip(actions, actions):
+            if values.get('post_move_sequence'):
+                for fiscalyear in fiscalyears:
+                    if (fiscalyear.post_move_sequence
+                            and fiscalyear.post_move_sequence.id !=
+                            values['post_move_sequence']):
+                        cls.raise_user_error('change_post_move_sequence', (
+                                fiscalyear.rec_name,))
+            values = values.copy()
+            if 'periods' in values:
+                operator = ['delete', 'remove', 'create', 'write', 'add']
+                values['periods'].sort(lambda x, y:
+                    cmp(operator.index(x[0]), operator.index(y[0])))
+            args.extend((fiscalyears, values))
+        super(FiscalYear, cls).write(*args)
 
     @classmethod
     def delete(cls, fiscalyears):
@@ -150,26 +166,28 @@ class FiscalYear(ModelSQL, ModelView):
         Create periods for the fiscal years with month interval
         '''
         Period = Pool().get('account.period')
+        to_create = []
         for fiscalyear in fiscalyears:
             period_start_date = fiscalyear.start_date
             while period_start_date < fiscalyear.end_date:
                 period_end_date = period_start_date + \
-                        relativedelta(months=interval - 1) + \
-                        relativedelta(day=31)
+                    relativedelta(months=interval - 1) + \
+                    relativedelta(day=31)
                 if period_end_date > fiscalyear.end_date:
                     period_end_date = fiscalyear.end_date
                 name = datetime_strftime(period_start_date, '%Y-%m')
                 if name != datetime_strftime(period_end_date, '%Y-%m'):
                     name += ' - ' + datetime_strftime(period_end_date, '%Y-%m')
-                Period.create({
+                to_create.append({
                     'name': name,
                     'start_date': period_start_date,
                     'end_date': period_end_date,
                     'fiscalyear': fiscalyear.id,
-                    'post_move_sequence': fiscalyear.post_move_sequence.id,
                     'type': 'standard',
                     })
                 period_start_date = period_end_date + relativedelta(days=1)
+        if to_create:
+            Period.create(to_create)
 
     @classmethod
     @ModelView.button
@@ -187,7 +205,9 @@ class FiscalYear(ModelSQL, ModelView):
         If exception is set the function will raise an exception
             if any fiscal year is found.
         '''
-        Date = Pool().get('ir.date')
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        Date = pool.get('ir.date')
 
         if not date:
             date = Date.today()
@@ -198,7 +218,13 @@ class FiscalYear(ModelSQL, ModelView):
             ], order=[('start_date', 'DESC')], limit=1)
         if not fiscalyears:
             if exception:
-                cls.raise_user_error('no_fiscalyear_date')
+                language = Transaction().language
+                languages = Lang.search([('code', '=', language)])
+                if not languages:
+                    languages = Lang.search([('code', '=', 'en_US')])
+                language, = languages
+                formatted = Lang.strftime(date, language.code, language.date)
+                cls.raise_user_error('no_fiscalyear_date', (formatted,))
             else:
                 return None
         return fiscalyears[0].id
@@ -217,12 +243,12 @@ class FiscalYear(ModelSQL, ModelView):
                 self.raise_user_error('account_balance_not_zero',
                         error_args=(account.rec_name,))
         else:
-            Deferral.create({
-                'account': account.id,
-                'fiscalyear': self.id,
-                'debit': account.debit,
-                'credit': account.credit,
-                })
+            Deferral.create([{
+                        'account': account.id,
+                        'fiscalyear': self.id,
+                        'debit': account.debit,
+                        'credit': account.credit,
+                        }])
 
     @classmethod
     @ModelView.button
@@ -236,11 +262,11 @@ class FiscalYear(ModelSQL, ModelView):
 
         for fiscalyear in fiscalyears:
             if cls.search([
-                ('end_date', '<=', fiscalyear.start_date),
-                ('state', '=', 'open'),
-                ('company', '=', fiscalyear.company.id),
-                ]):
-                cls.raise_user_error('close_error')
+                        ('end_date', '<=', fiscalyear.start_date),
+                        ('state', '=', 'open'),
+                        ('company', '=', fiscalyear.company.id),
+                        ]):
+                cls.raise_user_error('close_error', (fiscalyear.rec_name,))
 
             #First close the fiscalyear to be sure
             #it will not have new period created between.
@@ -270,10 +296,10 @@ class FiscalYear(ModelSQL, ModelView):
 
         for fiscalyear in fiscalyears:
             if cls.search([
-                ('start_date', '>=', fiscalyear.end_date),
-                ('state', '=', 'close'),
-                ('company', '=', fiscalyear.company.id),
-                ]):
+                        ('start_date', '>=', fiscalyear.end_date),
+                        ('state', '=', 'close'),
+                        ('company', '=', fiscalyear.company.id),
+                        ]):
                 cls.raise_user_error('reopen_error')
 
             deferrals = Deferral.search([
@@ -285,6 +311,126 @@ class FiscalYear(ModelSQL, ModelView):
                 'state': 'open',
                 })
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return ['OR',
+            ('code',) + tuple(clause[1:]),
+            (cls._rec_name,) + tuple(clause[1:]),
+            ]
+
+
+class BalanceNonDeferralStart(ModelView):
+    'Balance Non-Deferral'
+    __name__ = 'account.fiscalyear.balance_non_deferral.start'
+    fiscalyear = fields.Many2One('account.fiscalyear', 'Fiscal Year',
+        required=True, domain=[('state', '=', 'open')])
+    journal = fields.Many2One('account.journal', 'Journal', required=True,
+        domain=[
+            ('type', '=', 'situation'),
+            ])
+    period = fields.Many2One('account.period', 'Period', required=True,
+        domain=[
+            ('fiscalyear', '=', Eval('fiscalyear')),
+            ('type', '=', 'adjustment'),
+            ],
+        depends=['fiscalyear'])
+    credit_account = fields.Many2One('account.account', 'Credit Account',
+        required=True,
+        domain=[
+            ('kind', '!=', 'view'),
+            ('company', '=', Eval('context', {}).get('company', -1)),
+            ('deferral', '=', True),
+            ])
+    debit_account = fields.Many2One('account.account', 'Debit Account',
+        required=True,
+        domain=[
+            ('kind', '!=', 'view'),
+            ('company', '=', Eval('context', {}).get('company', -1)),
+            ('deferral', '=', True),
+            ])
+
+
+class BalanceNonDeferral(Wizard):
+    'Balance Non-Deferral'
+    __name__ = 'account.fiscalyear.balance_non_deferral'
+    start = StateView('account.fiscalyear.balance_non_deferral.start',
+        'account.fiscalyear_balance_non_deferral_start_view_form', [
+            Button('Cancel', 'end', 'tryton-cancel'),
+            Button('Ok', 'balance', 'tryton-ok', default=True),
+            ])
+    balance = StateAction('account.act_move_line_form')
+
+    def get_move_line(self, account):
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        if account.company.currency.is_zero(account.balance):
+            return
+        line = Line()
+        line.account = account
+        if account.balance >= 0:
+            line.credit = abs(account.balance)
+            line.debit = 0
+        else:
+            line.credit = 0
+            line.debit = abs(account.balance)
+        return line
+
+    def get_counterpart_line(self, amount):
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        if self.start.fiscalyear.company.currency.is_zero(amount):
+            return
+        line = Line()
+        if amount >= 0:
+            line.credit = abs(amount)
+            line.debit = 0
+            line.account = self.start.credit_account
+        else:
+            line.credit = 0
+            line.debit = abs(amount)
+            line.account = self.start.debit_account
+        return line
+
+    def create_move(self):
+        pool = Pool()
+        Account = pool.get('account.account')
+        Move = pool.get('account.move')
+
+        with Transaction().set_context(fiscalyear=self.start.fiscalyear.id,
+                date=None, cumulate=False):
+            accounts = Account.search([
+                    ('company', '=', self.start.fiscalyear.company.id),
+                    ('deferral', '=', False),
+                    ('kind', '!=', 'view'),
+                    ])
+        lines = []
+        for account in accounts:
+            line = self.get_move_line(account)
+            if line:
+                lines.append(line)
+        if not lines:
+            return
+        amount = sum(l.debit - l.credit for l in lines)
+        counter_part_line = self.get_counterpart_line(amount)
+        if counter_part_line:
+            lines.append(counter_part_line)
+
+        move = Move()
+        move.period = self.start.period
+        move.journal = self.start.journal
+        move.date = self.start.period.start_date
+        move.origin = self.start.fiscalyear
+        move.lines = lines
+        move.save()
+        return move
+
+    def do_balance(self, action):
+        self.create_move()
+        action['pyson_domain'] = PYSONEncoder().encode([
+                ('origin', '=', str(self.start.fiscalyear)),
+                ])
+        return action, {}
+
 
 class CloseFiscalYearStart(ModelView):
     'Close Fiscal Year'
