@@ -2,22 +2,23 @@
 #this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 import datetime
-import time
 from itertools import groupby
 from operator import itemgetter
+from sql.aggregate import Sum
+from sql.conditionals import Coalesce
 
 from trytond.model import ModelView, ModelSQL, fields
 from trytond.wizard import Wizard, StateTransition, StateView, StateAction, \
     Button
 from trytond.report import Report
-from trytond.backend import TableHandler, FIELDS
-from trytond.pyson import Eval, PYSONEncoder
+from trytond import backend
+from trytond.pyson import Eval, Bool, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 from trytond.rpc import RPC
 from trytond.tools import reduce_ids
 
-__all__ = ['Move', 'Reconciliation', 'Line', 'Move2', 'OpenJournalAsk',
+__all__ = ['Move', 'Reconciliation', 'Line', 'OpenJournalAsk',
     'OpenJournal', 'OpenAccount', 'ReconcileLinesWriteOff', 'ReconcileLines',
     'UnreconcileLinesStart', 'UnreconcileLines', 'OpenReconcileLinesStart',
     'OpenReconcileLines', 'FiscalYearLine', 'FiscalYear2',
@@ -46,7 +47,7 @@ class Move(ModelSQL, ModelView):
     journal = fields.Many2One('account.journal', 'Journal', required=True,
             states=_MOVE_STATES, depends=_MOVE_DEPENDS)
     date = fields.Date('Effective Date', required=True, states=_MOVE_STATES,
-        depends=_MOVE_DEPENDS, on_change_with=['period', 'journal', 'date'])
+        depends=_MOVE_DEPENDS)
     post_date = fields.Date('Post Date', readonly=True)
     description = fields.Char('Description', states=_MOVE_STATES,
         depends=_MOVE_DEPENDS)
@@ -79,10 +80,6 @@ class Move(ModelSQL, ModelView):
                     '"%(move)s" to draft in journal "%(journal)s".'),
                 'modify_posted_move': ('You can not modify move "%s" because '
                     'it is already posted.'),
-                'period_centralized_journal': ('Move "%(move)s" cannot be '
-                    'created because there is already a move in journal '
-                    '"%(journal)s" and you cannot create more than one move '
-                    'per period in a centralized journal.'),
                 'company_in_move': ('You can not create lines on accounts'
                     'of different companies in move "%s".'),
                 'date_outside_period': ('You can not create move "%(move)s" '
@@ -101,6 +98,7 @@ class Move(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
         table = TableHandler(cursor, cls, module_name)
 
@@ -141,6 +139,7 @@ class Move(ModelSQL, ModelView):
             return period.start_date
         return Date.today()
 
+    @fields.depends('period', 'journal', 'date')
     def on_change_with_date(self):
         Line = Pool().get('account.move.line')
         date = self.date
@@ -157,7 +156,7 @@ class Move(ModelSQL, ModelView):
     @classmethod
     def _get_origin(cls):
         'Return list of Model names for origin Reference'
-        return []
+        return ['account.fiscalyear']
 
     @classmethod
     def get_origin(cls):
@@ -172,23 +171,9 @@ class Move(ModelSQL, ModelView):
     def validate(cls, moves):
         super(Move, cls).validate(moves)
         for move in moves:
-            move.check_centralisation()
             move.check_company()
             move.check_date()
 
-    def check_centralisation(self):
-        if self.journal.centralised:
-            moves = self.search([
-                    ('period', '=', self.period.id),
-                    ('journal', '=', self.journal.id),
-                    ('state', '!=', 'posted'),
-                    ], limit=2)
-            if len(moves) > 1:
-                self.raise_user_error('period_centralized_journal', {
-                        'move': self.rec_name,
-                        'journal': self.journal.rec_name,
-                        })
-
     def check_company(self):
         company_id = -1
         for line in self.lines:
@@ -200,7 +185,9 @@ class Move(ModelSQL, ModelView):
     def check_date(self):
         if (self.date < self.period.start_date
                 or self.date > self.period.end_date):
-            self.raise_user_error('date_outside_period', (self.rec_name,))
+            self.raise_user_error('date_outside_period', {
+                        'move': self.rec_name,
+                        })
 
     @classmethod
     def check_modify(cls, moves):
@@ -211,27 +198,31 @@ class Move(ModelSQL, ModelView):
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        moves = cls.search(['OR',
-                ('post_number',) + tuple(clause[1:]),
-                (cls._rec_name,) + tuple(clause[1:]),
-                ])
-        return [('id', 'in', [m.id for m in moves])]
+        return ['OR',
+            ('post_number',) + tuple(clause[1:]),
+            (cls._rec_name,) + tuple(clause[1:]),
+            ]
 
     @classmethod
-    def write(cls, moves, vals):
-        keys = vals.keys()
-        for key in cls._check_modify_exclude:
-            if key in keys:
-                keys.remove(key)
-        if len(keys):
-            cls.check_modify(moves)
-        super(Move, cls).write(moves, vals)
-        cls.validate_move(moves)
+    def write(cls, *args):
+        actions = iter(args)
+        all_moves = []
+        args = []
+        for moves, values in zip(actions, actions):
+            keys = values.keys()
+            for key in cls._check_modify_exclude:
+                if key in keys:
+                    keys.remove(key)
+            if len(keys):
+                cls.check_modify(moves)
+            args.extend((moves, values))
+            all_moves.extend(moves)
+        super(Move, cls).write(*args)
+        cls.validate_move(all_moves)
 
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        MoveLine = pool.get('account.move.line')
         Sequence = pool.get('ir.sequence')
         Journal = pool.get('account.journal')
 
@@ -245,15 +236,6 @@ class Move(ModelSQL, ModelView):
                     vals['number'] = Sequence.get_id(journal.sequence.id)
 
         moves = super(Move, cls).create(vlist)
-        for move in moves:
-            if move.journal.centralised:
-                line, = MoveLine.create([{
-                            'account': move.journal.credit_account.id,
-                            'move': move.id,
-                            }])
-                cls.write([move], {
-                        'centralised_line': line.id,
-                        })
         cls.validate_move(moves)
         return moves
 
@@ -289,11 +271,12 @@ class Move(ModelSQL, ModelView):
     @classmethod
     def validate_move(cls, moves):
         '''
-        Validate balanced move and centralise it if in centralised journal
+        Validate balanced move
         '''
         pool = Pool()
         MoveLine = pool.get('account.move.line')
         User = pool.get('res.user')
+        line = MoveLine.__table__()
 
         cursor = Transaction().cursor
 
@@ -308,18 +291,17 @@ class Move(ModelSQL, ModelView):
         for i in range(0, len(moves), cursor.IN_MAX):
             sub_moves = moves[i:i + cursor.IN_MAX]
             sub_move_ids = [m.id for m in sub_moves]
-            red_sql, red_ids = reduce_ids('move', sub_move_ids)
+            red_sql = reduce_ids(line.move, sub_move_ids)
 
-            cursor.execute('SELECT move, SUM(debit - credit) '
-                'FROM "' + MoveLine._table + '" '
-                'WHERE ' + red_sql + ' '
-                'GROUP BY move', red_ids)
+            cursor.execute(*line.select(line.move,
+                    Sum(line.debit - line.credit),
+                    where=red_sql,
+                    group_by=line.move))
             amounts.update(dict(cursor.fetchall()))
 
-            cursor.execute('SELECT move, id '
-                'FROM "' + MoveLine._table + '" '
-                'WHERE ' + red_sql + ' AND state = %s '
-                'ORDER BY move', red_ids + ['draft'])
+            cursor.execute(*line.select(line.move, line.id,
+                    where=red_sql & (line.state == 'draft'),
+                    order_by=line.move))
             move2draft_lines.update(dict((k, (j[1] for j in g))
                     for k, g in groupby(cursor.fetchall(), itemgetter(0))))
 
@@ -335,39 +317,7 @@ class Move(ModelSQL, ModelView):
             draft_lines = MoveLine.browse(
                 list(move2draft_lines.get(move.id, [])))
             if not company.currency.is_zero(amount):
-                if not move.journal.centralised:
-                    draft_moves.append(move.id)
-                else:
-                    if not move.centralised_line:
-                        centralised_amount = - amount
-                    else:
-                        centralised_amount = move.centralised_line.debit \
-                            - move.centralised_line.credit \
-                            - amount
-                    if centralised_amount >= Decimal('0.0'):
-                        debit = centralised_amount
-                        credit = Decimal('0.0')
-                        account_id = move.journal.debit_account.id
-                    else:
-                        debit = Decimal('0.0')
-                        credit = - centralised_amount
-                        account_id = move.journal.credit_account.id
-                    if not move.centralised_line:
-                        centralised_line, = MoveLine.create([{
-                                    'debit': debit,
-                                    'credit': credit,
-                                    'account': account_id,
-                                    'move': move.id,
-                                    }])
-                        cls.write([move], {
-                            'centralised_line': centralised_line.id,
-                            })
-                    else:
-                        MoveLine.write([move.centralised_line], {
-                                'debit': debit,
-                                'credit': credit,
-                                'account': account_id,
-                                })
+                draft_moves.append(move.id)
                 continue
             if not draft_lines:
                 continue
@@ -379,11 +329,12 @@ class Move(ModelSQL, ModelView):
             if move_ids:
                 for i in range(0, len(move_ids), cursor.IN_MAX):
                     sub_ids = move_ids[i:i + cursor.IN_MAX]
-                    red_sql, red_ids = reduce_ids('move', sub_ids)
+                    red_sql = reduce_ids(line.move, sub_ids)
                     # Use SQL to prevent double validate loop
-                    cursor.execute('UPDATE "' + MoveLine._table + '" '
-                        'SET state = %s '
-                        'WHERE ' + red_sql, [state] + red_ids)
+                    cursor.execute(*line.update(
+                            columns=[line.state],
+                            values=[state],
+                            where=red_sql))
 
     @classmethod
     @ModelView.button
@@ -391,6 +342,7 @@ class Move(ModelSQL, ModelView):
         pool = Pool()
         Sequence = pool.get('ir.sequence')
         Date = pool.get('ir.date')
+        Line = pool.get('account.move.line')
 
         for move in moves:
             amount = Decimal('0.0')
@@ -413,6 +365,14 @@ class Move(ModelSQL, ModelView):
                     move.period.post_move_sequence_used.id)
             cls.write([move], values)
 
+            keyfunc = lambda l: (l.party, l.account)
+            to_reconcile = [l for l in move.lines
+                if ((l.debit == l.credit == Decimal('0'))
+                    and l.account.reconcile)]
+            to_reconcile = sorted(to_reconcile, key=keyfunc)
+            for _, zero_lines in groupby(to_reconcile, keyfunc):
+                Line.reconcile(list(zero_lines))
+
     @classmethod
     @ModelView.button
     def draft(cls, moves):
@@ -449,7 +409,8 @@ class Reconciliation(ModelSQL, ModelView):
                 'reconciliation_different_accounts': ('You can not reconcile '
                     'line "%(line)s" because it\'s account "%(account1)s" is '
                     'different from "%(account2)s".'),
-                'reconciliation_account_no_reconcile': ('You can not reconcile '
+                'reconciliation_account_no_reconcile': (
+                    'You can not reconcile '
                     'line "%(line)s" because it\'s account "%(account)s" is '
                     'configured as not reconcilable.'),
                 'reconciliation_different_parties': ('You can not reconcile '
@@ -472,7 +433,7 @@ class Reconciliation(ModelSQL, ModelView):
         return super(Reconciliation, cls).create(vlist)
 
     @classmethod
-    def write(cls, moves, vals):
+    def write(cls, moves, values, *args):
         cls.raise_user_error('modify')
 
     @classmethod
@@ -537,19 +498,13 @@ class Line(ModelSQL, ModelView):
     __name__ = 'account.move.line'
     debit = fields.Numeric('Debit', digits=(16, Eval('currency_digits', 2)),
         required=True,
-        on_change=['account', 'debit', 'credit', 'tax_lines', 'journal',
-            'move'],
         depends=['currency_digits', 'credit', 'tax_lines', 'journal'])
     credit = fields.Numeric('Credit', digits=(16, Eval('currency_digits', 2)),
         required=True,
-        on_change=['account', 'debit', 'credit', 'tax_lines', 'journal',
-            'move'],
         depends=['currency_digits', 'debit', 'tax_lines', 'journal'])
     account = fields.Many2One('account.account', 'Account', required=True,
             domain=[('kind', '!=', 'view')],
-            select=True,
-            on_change=['account', 'debit', 'credit', 'tax_lines',
-                'journal', 'move'])
+            select=True)
     move = fields.Many2One('account.move', 'Move', select=True, required=True,
         states={
             'required': False,
@@ -569,16 +524,23 @@ class Line(ModelSQL, ModelView):
             selection='get_origin'),
         'get_move_field', searcher='search_move_field')
     description = fields.Char('Description')
+    move_description = fields.Function(fields.Char('Move Description'),
+        'get_move_field', setter='set_move_field',
+        searcher='search_move_field')
     amount_second_currency = fields.Numeric('Amount Second Currency',
             digits=(16, Eval('second_currency_digits', 2)),
             help='The amount expressed in a second currency',
-            depends=['second_currency_digits'])
+        states={
+            'required': Bool(Eval('second_currency')),
+            },
+        depends=['second_currency_digits', 'second_currency'])
     second_currency = fields.Many2One('currency.currency', 'Second Currency',
-            help='The second currency')
-    party = fields.Many2One('party.party', 'Party',
-            on_change=['move', 'party', 'account', 'debit', 'credit',
-                'journal'], select=True, depends=['debit', 'credit', 'account',
-                    'journal'])
+            help='The second currency',
+        states={
+            'required': Bool(Eval('amount_second_currency')),
+            },
+        depends=['amount_second_currency'])
+    party = fields.Many2One('party.party', 'Party', select=True)
     maturity_date = fields.Date('Maturity Date',
         help='This field is used for payable and receivable lines. \n'
         'You can put the limit date for the payment.')
@@ -601,6 +563,7 @@ class Line(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Line, cls).__setup__()
+        cls._check_modify_exclude = ['reconciliation']
         cls._sql_constraints += [
             ('credit_debit',
                 'CHECK(credit * debit = 0.0)',
@@ -616,19 +579,20 @@ class Line(ModelSQL, ModelView):
                     'add/modify lines in closed journal period "%s".'),
                 'modify_posted_move': ('You can not modify lines of move "%s" '
                     'because it is already posted.'),
-                'modify_reconciled': ('You can not modify line "%s" because it '
-                    'is reconciled.'),
+                'modify_reconciled': ('You can not modify line "%s" because '
+                    'it is reconciled.'),
                 'no_journal': ('Move line cannot be created because there is '
                     'no journal defined.'),
                 'move_view_account': ('You can not create a move line with '
                     'account "%s" because it is a view account.'),
-                'move_inactive_account': ('You can not create a move line with '
-                    'account "%s" because it is inactive.'),
+                'move_inactive_account': ('You can not create a move line '
+                    'with account "%s" because it is inactive.'),
                 'already_reconciled': 'Line "%s" (%d) already reconciled.',
                 })
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
         table = TableHandler(cursor, cls, module_name)
 
@@ -645,8 +609,10 @@ class Line(ModelSQL, ModelView):
         # Migration from 1.2
         table.not_null_action('blocked', action='remove')
 
-        # Migration from 2.4: remove name
+        # Migration from 2.4: remove name, active
         table.not_null_action('name', action='remove')
+        table.not_null_action('active', action='remove')
+        table.index_action('active', action='remove')
 
     @classmethod
     def default_date(cls):
@@ -690,14 +656,14 @@ class Line(ModelSQL, ModelView):
         return Decimal(0)
 
     @classmethod
-    def default_get(cls, fields, with_rec_name=True):
+    def default_get(cls, fields, with_rec_name=True, with_on_change=True):
         pool = Pool()
         Move = pool.get('account.move')
         Tax = pool.get('account.tax')
         Account = pool.get('account.account')
         TaxCode = pool.get('account.tax.code')
         values = super(Line, cls).default_get(fields,
-                with_rec_name=with_rec_name)
+                with_rec_name=with_rec_name, with_on_change=with_on_change)
 
         if 'move' not in fields:
             #Not manual entry
@@ -879,6 +845,8 @@ class Line(ModelSQL, ModelView):
         Move = Pool().get('account.move')
         return Move.get_origin()
 
+    @fields.depends('account', 'debit', 'credit', 'tax_lines', 'journal',
+        'move')
     def on_change_debit(self):
         changes = {}
         Journal = Pool().get('account.journal')
@@ -892,6 +860,8 @@ class Line(ModelSQL, ModelView):
             changes['credit'] = Decimal('0.0')
         return changes
 
+    @fields.depends('account', 'debit', 'credit', 'tax_lines', 'journal',
+        'move')
     def on_change_credit(self):
         changes = {}
         Journal = Pool().get('account.journal')
@@ -905,6 +875,8 @@ class Line(ModelSQL, ModelView):
             changes['debit'] = Decimal('0.0')
         return changes
 
+    @fields.depends('account', 'debit', 'credit', 'tax_lines', 'journal',
+        'move')
     def on_change_account(self):
         Journal = Pool().get('account.journal')
 
@@ -975,6 +947,7 @@ class Line(ModelSQL, ModelView):
                     res.setdefault('add', []).append(value)
         return res
 
+    @fields.depends('move', 'party', 'account', 'debit', 'credit', 'journal')
     def on_change_party(self):
         Journal = Pool().get('account.journal')
         cursor = Transaction().cursor
@@ -987,18 +960,15 @@ class Line(ModelSQL, ModelView):
             return res
 
         if self.party and (not self.debit) and (not self.credit):
-            type_name = FIELDS[self.__class__.debit._type].sql_type(
-                self.__class__.debit)[0]
-            query = ('SELECT '
-                'CAST(COALESCE(SUM('
-                        '(COALESCE(debit, 0) - COALESCE(credit, 0))'
-                    '), 0) AS ' + type_name + ') '
-                'FROM account_move_line '
-                'WHERE reconciliation IS NULL '
-                    'AND party = %s '
-                    'AND account = %s')
-            cursor.execute(query,
-                (self.party.id, self.party.account_receivable.id))
+            type_name = self.__class__.debit.sql_type().base
+            table = self.__table__()
+            column = Coalesce(Sum(Coalesce(table.debit, 0)
+                    - Coalesce(table.credit, 0)), 0).cast(type_name)
+            where = ((table.reconciliation == None)
+                & (table.party == self.party.id))
+            cursor.execute(*table.select(column,
+                    where=where
+                    & (table.account == self.party.account_receivable.id)))
             amount = cursor.fetchone()[0]
             # SQLite uses float for SUM
             if not isinstance(amount, Decimal):
@@ -1016,8 +986,9 @@ class Line(ModelSQL, ModelView):
                 res['account.rec_name'] = \
                     self.party.account_receivable.rec_name
             else:
-                cursor.execute(query,
-                    (self.party.id, self.party.account_payable.id))
+                cursor.execute(*table.select(column,
+                        where=where
+                        & (table.account == self.party.account_payable.id)))
                 amount = cursor.fetchone()[0]
                 if not self.party.account_payable.currency.is_zero(amount):
                     if amount > Decimal('0.0'):
@@ -1075,9 +1046,9 @@ class Line(ModelSQL, ModelView):
         return res
 
     def get_move_field(self, name):
-        if name == 'move_state':
-            name = 'state'
-        if name in ('date', 'state'):
+        if name.startswith('move_'):
+            name = name[5:]
+        if name in ('date', 'state', 'description'):
             return getattr(self.move, name)
         elif name == 'origin':
             origin = getattr(self.move, name)
@@ -1089,8 +1060,8 @@ class Line(ModelSQL, ModelView):
 
     @classmethod
     def set_move_field(cls, lines, name, value):
-        if name == 'move_state':
-            name = 'state'
+        if name.startswith('move_'):
+            name = name[5:]
         if not value:
             return
         Move = Pool().get('account.move')
@@ -1100,21 +1071,46 @@ class Line(ModelSQL, ModelView):
 
     @classmethod
     def search_move_field(cls, name, clause):
-        if name == 'move_state':
-            name = 'state'
+        if name.startswith('move_'):
+            name = name[5:]
         return [('move.' + name,) + tuple(clause[1:])]
 
+    def _order_move_field(name):
+        def order_field(tables):
+            pool = Pool()
+            Move = pool.get('account.move')
+            field = Move._fields[name]
+            table, _ = tables[None]
+            move_tables = tables.get('move')
+            if move_tables is None:
+                move = Move.__table__()
+                move_tables = {
+                    None: (move, move.id == table.move),
+                    }
+                tables['move'] = move_tables
+            return field.convert_order(name, move_tables, Move)
+        return staticmethod(order_field)
+    order_journal = _order_move_field('journal')
+    order_period = _order_move_field('period')
+    order_date = _order_move_field('date')
+    order_origin = _order_move_field('origin')
+    order_move_state = _order_move_field('state')
+
     @classmethod
-    def query_get(cls, obj='l'):
+    def query_get(cls, table):
         '''
         Return SQL clause and fiscal years for account move line
         depending of the context.
-        obj is the SQL alias of account_move_line in the query
+        table is the SQL instance of account.move.line table
         '''
-        FiscalYear = Pool().get('account.fiscalyear')
+        pool = Pool()
+        FiscalYear = pool.get('account.fiscalyear')
+        Move = pool.get('account.move')
+        Period = pool.get('account.period')
+        move = Move.__table__()
+        period = Period.__table__()
 
         if Transaction().context.get('date'):
-            time.strptime(str(Transaction().context['date']), '%Y-%m-%d')
             fiscalyears = FiscalYear.search([
                     ('start_date', '<=', Transaction().context['date']),
                     ('end_date', '>=', Transaction().context['date']),
@@ -1123,83 +1119,71 @@ class Line(ModelSQL, ModelView):
             fiscalyear_id = fiscalyears and fiscalyears[0].id or 0
 
             if Transaction().context.get('posted'):
-                return (obj + '.state != \'draft\' '
-                        'AND ' + obj + '.move IN ('
-                            'SELECT m.id FROM account_move AS m, '
-                                'account_period AS p '
-                                'WHERE m.period = p.id '
-                                    'AND p.fiscalyear = ' +
-                                        str(fiscalyear_id) + ' '
-                                    'AND m.date <= date(\'' +
-                                        str(Transaction().context['date']) +
-                                        '\') '
-                                    'AND m.state = \'posted\' '
-                            ')', [f.id for f in fiscalyears])
+                return ((table.state != 'draft')
+                    & table.move.in_(move.join(period,
+                            condition=move.period == period.id
+                            ).select(move.id,
+                            where=(period.fiscalyear == fiscalyear_id)
+                            & (move.date <= Transaction().context['date'])
+                            & (move.state == 'posted'))),
+                    [f.id for f in fiscalyears])
             else:
-                return (obj + '.state != \'draft\' '
-                        'AND ' + obj + '.move IN ('
-                            'SELECT m.id FROM account_move AS m, '
-                                'account_period AS p '
-                                'WHERE m.period = p.id '
-                                    'AND p.fiscalyear = ' +
-                                        str(fiscalyear_id) + ' '
-                                    'AND m.date <= date(\'' +
-                                        str(Transaction().context['date']) +
-                                        '\')'
-                            ')', [f.id for f in fiscalyears])
+                return ((table.state != 'draft')
+                    & table.move.in_(move.join(period,
+                            condition=move.period == period.id
+                            ).select(move.id,
+                            where=(period.fiscalyear == fiscalyear_id)
+                            & (move.date <= Transaction().context['date']))),
+                    [f.id for f in fiscalyears])
 
         if Transaction().context.get('periods'):
             if Transaction().context.get('fiscalyear'):
-                fiscalyear_ids = [int(Transaction().context['fiscalyear'])]
+                fiscalyear_ids = [Transaction().context['fiscalyear']]
             else:
                 fiscalyear_ids = []
-            ids = ','.join(
-                    str(int(x)) for x in Transaction().context['periods'])
             if Transaction().context.get('posted'):
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN (' + ids + ') '
-                            'AND state = \'posted\' '
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                            move.select(move.id,
+                                where=move.period.in_(
+                                    Transaction().context['periods'])
+                                & (move.state == 'posted'))),
+                    fiscalyear_ids)
             else:
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN (' + ids + ')'
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                        move.select(move.id,
+                            where=move.period.in_(
+                                Transaction().context['periods']))),
+                    fiscalyear_ids)
         else:
             if not Transaction().context.get('fiscalyear'):
                 fiscalyears = FiscalYear.search([
                     ('state', '=', 'open'),
                     ])
-                fiscalyear_ids = [f.id for f in fiscalyears]
-                fiscalyear_clause = (','.join(
-                        str(f.id) for f in fiscalyears)) or '0'
+                fiscalyear_ids = [f.id for f in fiscalyears] or [0]
             else:
-                fiscalyear_ids = [int(Transaction().context.get('fiscalyear'))]
-                fiscalyear_clause = '%s' % int(
-                        Transaction().context.get('fiscalyear'))
+                fiscalyear_ids = [Transaction().context.get('fiscalyear')]
 
             if Transaction().context.get('posted'):
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN ('
-                            'SELECT id FROM account_period '
-                            'WHERE fiscalyear IN (' + fiscalyear_clause + ')'
-                            ') '
-                            'AND state = \'posted\' '
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                        move.select(move.id,
+                            where=move.period.in_(
+                                period.select(period.id,
+                                    where=period.fiscalyear.in_(
+                                        fiscalyear_ids)))
+                            & (move.state == 'posted'))),
+                    fiscalyear_ids)
             else:
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN ('
-                            'SELECT id FROM account_period '
-                            'WHERE fiscalyear IN (' + fiscalyear_clause + ')'
-                            ')'
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                        move.select(move.id,
+                            where=move.period.in_(
+                                period.select(period.id,
+                                    where=period.fiscalyear.in_(
+                                        fiscalyear_ids))))),
+                    fiscalyear_ids)
 
     @classmethod
     def on_write(cls, lines):
@@ -1270,17 +1254,24 @@ class Line(ModelSQL, ModelView):
         Move.validate_move(moves)
 
     @classmethod
-    def write(cls, lines, vals):
+    def write(cls, *args):
         Move = Pool().get('account.move')
 
-        if len(vals) > 1 or 'reconciliation' not in vals:
-            cls.check_modify(lines)
-        moves = [x.move for x in lines]
-        super(Line, cls).write(lines, vals)
+        actions = iter(args)
+        args = []
+        moves = []
+        all_lines = []
+        for lines, values in zip(actions, actions):
+            if any(k not in cls._check_modify_exclude for k in values):
+                cls.check_modify(lines)
+            moves.extend((x.move for x in lines))
+            all_lines.extend(lines)
+            args.extend((lines, values))
+
+        super(Line, cls).write(*args)
 
         Transaction().timestamp = {}
-
-        Move.validate_move(list(set(l.move for l in lines) | set(moves)))
+        Move.validate_move(list(set(l.move for l in all_lines) | set(moves)))
 
     @classmethod
     def create(cls, vlist):
@@ -1295,15 +1286,6 @@ class Line(ModelSQL, ModelView):
                 if not journal_id:
                     cls.raise_user_error('no_journal')
                 journal = Journal(journal_id)
-                if journal.centralised:
-                    moves = Move.search([
-                            ('period', '=', vals.get('period')
-                                or Transaction().context.get('period')),
-                            ('journal', '=', journal_id),
-                            ('state', '!=', 'posted'),
-                            ], limit=1)
-                    if moves:
-                        vals['move'] = moves[0].id
                 if not vals.get('move'):
                     vals['move'] = Move.create([{
                                 'period': (vals.get('period')
@@ -1387,7 +1369,8 @@ class Line(ModelSQL, ModelView):
         return result
 
     @classmethod
-    def reconcile(cls, lines, journal=None, date=None, account=None):
+    def reconcile(cls, lines, journal=None, date=None, account=None,
+            description=None):
         pool = Pool()
         Move = pool.get('account.move')
         Reconciliation = pool.get('account.move.reconciliation')
@@ -1400,21 +1383,27 @@ class Line(ModelSQL, ModelView):
                         error_args=(line.move.number, line.id,))
 
         lines = lines[:]
+        reconcile_account = None
+        amount = Decimal('0.0')
+        for line in lines:
+            amount += line.debit - line.credit
+            if not reconcile_account:
+                reconcile_account = line.account
+        amount = reconcile_account.currency.round(amount)
+        if not account and journal:
+            if amount >= 0:
+                account = journal.debit_account
+            else:
+                account = journal.credit_account
         if journal and account:
             if not date:
                 date = Date.today()
-            reconcile_account = None
-            amount = Decimal('0.0')
-            for line in lines:
-                amount += line.debit - line.credit
-                if not reconcile_account:
-                    reconcile_account = line.account
-            amount = reconcile_account.currency.round(amount)
             period_id = Period.find(reconcile_account.company.id, date=date)
             move, = Move.create([{
                         'journal': journal.id,
                         'period': period_id,
                         'date': date,
+                        'description': description,
                         'lines': [
                             ('create', [{
                                         'account': reconcile_account.id,
@@ -1444,12 +1433,6 @@ class Line(ModelSQL, ModelView):
                     }])[0]
 
 
-class Move2:
-    __name__ = 'account.move'
-    centralised_line = fields.Many2One('account.move.line', 'Centralised Line',
-            readonly=True)
-
-
 class OpenJournalAsk(ModelView):
     'Open Journal Ask'
     __name__ = 'account.move.open_journal.ask'
@@ -1575,10 +1558,15 @@ class OpenAccount(Wizard):
 class ReconcileLinesWriteOff(ModelView):
     'Reconcile Lines Write-Off'
     __name__ = 'account.move.reconcile_lines.writeoff'
-    journal = fields.Many2One('account.journal', 'Journal', required=True)
+    journal = fields.Many2One('account.journal', 'Journal', required=True,
+        domain=[
+            ('type', '=', 'write-off'),
+            ])
     date = fields.Date('Date', required=True)
-    account = fields.Many2One('account.account', 'Account', required=True,
-            domain=[('kind', '!=', 'view')])
+    amount = fields.Numeric('Amount', digits=(16, Eval('currency_digits', 2)),
+        readonly=True, depends=['currency_digits'])
+    currency_digits = fields.Integer('Currency Digits', readonly=True)
+    description = fields.Char('Description')
 
     @staticmethod
     def default_date():
@@ -1597,7 +1585,8 @@ class ReconcileLines(Wizard):
             ])
     reconcile = StateTransition()
 
-    def transition_start(self):
+    def get_writeoff(self):
+        "Return writeoff amount and company"
         Line = Pool().get('account.move.line')
 
         company = None
@@ -1606,20 +1595,31 @@ class ReconcileLines(Wizard):
             amount += line.debit - line.credit
             if not company:
                 company = line.account.company
+        return amount, company
+
+    def transition_start(self):
+        amount, company = self.get_writeoff()
         if not company:
             return 'end'
         if company.currency.is_zero(amount):
             return 'reconcile'
         return 'writeoff'
 
+    def default_writeoff(self, fields):
+        amount, company = self.get_writeoff()
+        return {
+            'amount': amount,
+            'currency_digits': company.currency.digits,
+            }
+
     def transition_reconcile(self):
         Line = Pool().get('account.move.line')
 
         journal = getattr(self.writeoff, 'journal', None)
         date = getattr(self.writeoff, 'date', None)
-        account = getattr(self.writeoff, 'account', None)
+        description = getattr(self.writeoff, 'description', None)
         Line.reconcile(Line.browse(Transaction().context['active_ids']),
-            journal, date, account)
+            journal=journal, date=date, description=description)
         return 'end'
 
 
@@ -1664,10 +1664,18 @@ class OpenReconcileLines(Wizard):
     open_ = StateAction('account.act_move_line_form')
 
     def do_open_(self, action):
-        action['pyson_domain'] = PYSONEncoder().encode([
+        encoder = PYSONEncoder()
+        action['pyson_domain'] = encoder.encode([
             ('account', '=', self.start.account.id),
             ('reconciliation', '=', None),
             ])
+        if self.start.account.kind == 'receivable':
+            order = [('credit', 'DESC'), ('id', 'DESC')]
+        elif self.start.account.kind == 'payable':
+            order = [('debit', 'DESC'), ('id', 'DESC')]
+        else:
+            order = None
+        action['pyson_order'] = encoder.encode(order)
         return action, {}
 
 
